/**
 * External dependencies
 */
import { isEmpty } from 'lodash';
import { render, fireEvent } from '@testing-library/react';

/**
 * Internal dependencies
 */
import { Slot, Fill, Provider } from '../';

/**
 * WordPress dependencies
 */
import { Component } from '@wordpress/element';

class Filler extends Component {
	constructor() {
		super( ...arguments );

		this.state = {
			num: 1,
		};
	}
	render() {
		return [
			<button
				key="1"
				type="button"
				onClick={ () => this.setState( { num: this.state.num + 1 } ) }
			/>,
			<Fill name={ this.props.name } key="2">
				{ this.props.text || this.state.num.toString() }
			</Fill>,
		];
	}
}

describe( 'Slot', () => {
	it( 'should render empty Fills', () => {
		const { container } = render(
			<Provider>
				<div>
					<Slot name="chicken" />
				</div>
				<Fill name="chicken" />
			</Provider>
		);

		expect( container ).toMatchSnapshot();
	} );

	it( 'should render a string Fill', () => {
		const { container } = render(
			<Provider>
				<div>
					<Slot name="chicken" />
				</div>
				<Fill name="chicken">content</Fill>
			</Provider>
		);

		expect( container ).toMatchSnapshot();
	} );

	it( 'should render a Fill containing an element', () => {
		const { container } = render(
			<Provider>
				<div>
					<Slot name="chicken" />
				</div>
				<Fill name="chicken">
					<span />
				</Fill>
			</Provider>
		);

		expect( container ).toMatchSnapshot();
	} );

	it( 'should render a Fill containing an array', () => {
		const { container } = render(
			<Provider>
				<div>
					<Slot name="chicken" />
				</div>
				<Fill name="chicken">
					{ [ <span key="1" />, <div key="2" />, 'text' ] }
				</Fill>
			</Provider>
		);

		expect( container ).toMatchSnapshot();
	} );

	it( 'calls the functions passed as the Slot’s fillProps in the Fill', () => {
		const onClose = jest.fn();

		const { getByText } = render(
			<Provider>
				<Slot name="chicken" fillProps={ { onClose } } />
				<Fill name="chicken">
					{ ( props ) => {
						return (
							<button onClick={ props.onClose }>Click me</button>
						);
					} }
				</Fill>
			</Provider>
		);

		fireEvent.click( getByText( 'Click me' ) );

		expect( onClose ).toHaveBeenCalledTimes( 1 );
	} );

	it( 'should render empty Fills without HTML wrapper when render props used', () => {
		const { container } = render(
			<Provider>
				<div>
					<Slot name="chicken">
						{ ( fills ) =>
							! isEmpty( fills ) && (
								<blockquote>{ fills }</blockquote>
							)
						}
					</Slot>
				</div>
				<Fill name="chicken" />
			</Provider>
		);

		expect( container ).toMatchSnapshot();
	} );

	it( 'should render a string Fill with HTML wrapper when render props used', () => {
		const { container } = render(
			<Provider>
				<div>
					<Slot name="chicken">
						{ ( fills ) =>
							fills && <blockquote>{ fills }</blockquote>
						}
					</Slot>
				</div>
				<Fill name="chicken">content</Fill>
			</Provider>
		);

		expect( container ).toMatchSnapshot();
	} );

	it( 'should re-render Slot when not bubbling virtually', () => {
		const { container, getByRole } = render(
			<Provider>
				<div>
					<Slot name="egg" />
				</div>
				<Filler name="egg" />
			</Provider>
		);

		expect( container ).toMatchSnapshot();

		fireEvent.click( getByRole( 'button' ) );

		expect( container ).toMatchSnapshot();
	} );

	it( 'should render in expected order', () => {
		const { container, rerender } = render(
			<Provider>
				<div key="slot">
					<Slot name="egg" />
				</div>
			</Provider>
		);

		rerender(
			<Provider>
				<div key="slot">
					<Slot name="egg" />
				</div>
				<Filler name="egg" key="first" text="first" />
				<Filler name="egg" key="second" text="second" />
			</Provider>
		);

		rerender(
			<Provider>
				<div key="slot">
					<Slot name="egg" />
				</div>
				<Filler name="egg" key="second" text="second" />
			</Provider>
		);

		rerender(
			<Provider>
				<div key="slot">
					<Slot name="egg" />
				</div>
				<Filler name="egg" key="first" text="first" />
				<Filler name="egg" key="second" text="second" />
			</Provider>
		);

		expect( container ).toMatchSnapshot();
	} );

	describe.each( [ false, true ] )(
		'bubblesVirtually %p',
		( bubblesVirtually ) => {
			it( 'should subsume another slot by the same name', () => {
				const { container, rerender } = render(
					<Provider>
						<div data-position="first">
							<Slot
								name="egg"
								bubblesVirtually={ bubblesVirtually }
							/>
						</div>
						<div data-position="second"></div>
						<Fill name="egg">Content</Fill>
					</Provider>
				);

				rerender(
					<Provider>
						<div data-position="first">
							<Slot
								name="egg"
								bubblesVirtually={ bubblesVirtually }
							/>
						</div>
						<div data-position="second">
							<Slot
								name="egg"
								bubblesVirtually={ bubblesVirtually }
							/>
						</div>
						<Fill name="egg">Content</Fill>
					</Provider>
				);

				expect( container ).toMatchSnapshot();

				rerender(
					<Provider>
						<div data-position="first"></div>
						<div data-position="second">
							<Slot
								name="egg"
								bubblesVirtually={ bubblesVirtually }
							/>
						</div>
						<Fill name="egg">Content</Fill>
					</Provider>
				);

				expect( container ).toMatchSnapshot();
			} );

			it( 'should unmount two slots with the same name', () => {
				const { rerender, container } = render(
					<Provider>
						<div data-position="first">
							<Slot
								name="egg"
								bubblesVirtually={ bubblesVirtually }
							/>
						</div>
						<div data-position="second">
							<Slot
								name="egg"
								bubblesVirtually={ bubblesVirtually }
							/>
						</div>
						<Fill name="egg">Content</Fill>
					</Provider>
				);
				rerender(
					<Provider>
						<div data-position="first">
							<Slot
								name="egg"
								bubblesVirtually={ bubblesVirtually }
							/>
						</div>
						<div data-position="second" />
						<Fill name="egg">Content</Fill>
					</Provider>
				);
				rerender(
					<Provider>
						<div data-position="first" />
						<div data-position="second" />
						<Fill name="egg">Content</Fill>
					</Provider>
				);
				expect( container ).toMatchSnapshot();
			} );
		}
	);
} );
